/*
wordle.c Wordle9511, ejemplo de juego sobrediseñado con TDAs.
Copyright (C) 2022 Sebastian Santisi <ssantisi@fi.uba.ar>
*/
#include "wordle.h"
#include "config.h"
#include "letra.h"
#include "interaccion.h"

#include <stdlib.h>
#include <string.h>


#define LETRAS_ALFABETO 26


// El juego consiste en una palabra a adivinar, un alfabeto que indica el
// estado de cada una de las letras según los intentos ya realizados, un vector
// intentos con los n_intentos ya realizados y procesados y, finalmente, un
// indicador de si el juego está ganado.
// El vector de intentos sólo tiene instancia de letras para los n_intentos
// realizados.
// El juego se considera ganado cuando el último intento fue palabra
// n_intentos no puede ser superior a INTENTOS_MAXIMOS, en ese caso el juego se
// considera perdido
struct wordle {
    char palabra[PALABRA_LONGITUD + 1];
    letra_t *alfabeto[LETRAS_ALFABETO];  // Ver vector.h
    letra_t *intentos[INTENTOS_MAXIMOS][PALABRA_LONGITUD];  // Ver vector.h
    size_t n_intentos;
    bool ganado;
};

wordle_t *wordle_crear(const char *palabra) {
    wordle_t *w = malloc(sizeof(wordle_t));
    if(w == NULL)
        return NULL;

    strcpy(w->palabra, palabra);
    w->n_intentos = 0;
    w->ganado = false;

    for(size_t i = 0; i < LETRAS_ALFABETO; i++) {
        w->alfabeto[i] = letra_crear('A' + i);
        if(w->alfabeto[i] == NULL) {
            while(i--)
                letra_destruir(w->alfabeto[i]);
            free(w);
            return NULL;
        }
    }

    return w;
}

void wordle_destruir(wordle_t *w) {
    for(size_t i = 0; i < LETRAS_ALFABETO; i++)
        letra_destruir(w->alfabeto[i]);

    for(size_t i = 0; i < w->n_intentos; i++)
        for(size_t j = 0; j < PALABRA_LONGITUD; j++)
            letra_destruir(w->intentos[i][j]);

    free(w);
}

size_t wordle_get_cantidad_intentos(const wordle_t *w) {
    return w->n_intentos;
}

bool wordle_esta_ganado(const wordle_t *w) {
    return w->ganado;
}

bool wordle_esta_perdido(const wordle_t *w) {
    return ! w->ganado && w->n_intentos == INTENTOS_MAXIMOS;
}

bool wordle_procesar_intento(wordle_t *w, const char *palabra) {
    if(w->ganado || w->n_intentos == INTENTOS_MAXIMOS)
        return false;

    for(size_t i = 0; i < PALABRA_LONGITUD; i++) {
        w->intentos[w->n_intentos][i] = letra_crear(palabra[i]);
        if(w->intentos[w->n_intentos][i] == NULL) {
            while(i--)
                letra_destruir(w->intentos[w->n_intentos][i]);
            return false;
        }
    }

    bool no_coincidentes[PALABRA_LONGITUD] = {false};
    char letras_no_encontradas[PALABRA_LONGITUD];
    size_t n_letras_no_encontradas = 0;

    // Verificamos coincidencia exacta:
    for(size_t i = 0; i < PALABRA_LONGITUD; i++) {
        if(palabra[i] == w->palabra[i]) {
            letra_set_estado(w->intentos[w->n_intentos][i], LETRA_POSICION_CORRECTA);
            letra_set_estado(w->alfabeto[palabra[i] - 'A'], LETRA_POSICION_CORRECTA);
        }
        else {
            no_coincidentes[i] = true;
            letras_no_encontradas[n_letras_no_encontradas++] = w->palabra[i];
        }
    }

    for(size_t i = 0; i < PALABRA_LONGITUD; i++) {
        if(! no_coincidentes[i])    // Esa letra ya estaba donde iba
            continue;

        size_t j;
        for(j = 0; j < n_letras_no_encontradas; j++) {
            if(palabra[i] == letras_no_encontradas[j] && letra_get_estado(w->intentos[w->n_intentos][i]) == LETRA_NO_ARRIESGADA) {
                // La letra está en una posición incorrecta
                letra_set_estado(w->intentos[w->n_intentos][i], LETRA_POSICION_INCORRECTA);
                if(letra_get_estado(w->alfabeto[palabra[i] - 'A']) != LETRA_POSICION_CORRECTA)
                    letra_set_estado(w->alfabeto[palabra[i] - 'A'], LETRA_POSICION_INCORRECTA);
                letras_no_encontradas[j] = letras_no_encontradas[--n_letras_no_encontradas];
                j--;    // Si no decrementamos j y era la última va a entrar al siguiente if, gracias GDB!
                break;
            }
        }
        if(j == n_letras_no_encontradas) { // La letra no está
            letra_set_estado(w->intentos[w->n_intentos][i], LETRA_NO_ACERTADA);
            if(letra_get_estado(w->alfabeto[palabra[i] - 'A']) == LETRA_NO_ARRIESGADA)
                letra_set_estado(w->alfabeto[palabra[i] - 'A'], LETRA_NO_ACERTADA);
        }
    }

    if(!strcmp(palabra, w->palabra)) {
        w->ganado = true;
    }

    w->n_intentos++;

    return true;
}

bool wordle_imprimir_intento(const wordle_t *w, size_t i) {
    if(i >= w->n_intentos)
        return false;

    imprimir_letras(w->intentos[i], PALABRA_LONGITUD);
    return true;
}

bool wordle_imprimir_alfabeto(const wordle_t *w) {
    imprimir_letras(w->alfabeto, LETRAS_ALFABETO);
    return true;
}

bool imprimir_tabla_intentos(const wordle_t *w) {
    letra_t *letras[PALABRA_LONGITUD];
    for(size_t i = 0; i < PALABRA_LONGITUD; i++) {
        letras[i] = letra_crear(' ');
        if(letras[i] == NULL) {
            while(i++)
                letra_destruir(letras[i]);
            return false;
        }
    }

    for(size_t i = 0; i < w->n_intentos; i++) {
        for(size_t j = 0; j < PALABRA_LONGITUD; j++)
            letra_set_estado(letras[j], letra_get_estado(w->intentos[i][j]));
        imprimir_letras(letras, PALABRA_LONGITUD);
    }

    for(size_t i = 0; i < PALABRA_LONGITUD; i++)
        letra_destruir(letras[i]);

    return true;
}
